/* simply include tin's autoconf'd header file */
#include "autoconf.h"
#define EBCDIC 0
#define POSIX_MALLOC_THRESHOLD 10
#define PCRE_EXPORT
#define MAX_NAME_SIZE 32
#define MAX_NAME_COUNT 10000
#define MAX_DUPLENGTH 30000
